#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import warnings
import math
import pandas as pd
import scipy.stats as spstat
from . import lpdensity_fn
from . import lpbwdensity_fn
import plotnine as pn

pd.options.mode.chained_assignment = None

warnings.filterwarnings("ignore")


def lpdensity(data, grid=None, bw=None, p=None, q=None, v=None,
	kernel='triangular', scale=None,
	massPoints=True, bwselect='mse-dpi',
	stdVar=True, regularize=True,nLocalMin=None, nUniqueMin=None,
	Cweights=None, Pweights=None):
	
	"""
	Parameters
	----------
	data : vector
		Numeric vector or one dimensional matrix/data frame, the raw data.
	grid : number or vector
		Numeric, specifies the grid of evaluation points. When set to default, grid points will be chosen as 0.05-0.95 percentiles of the data, with a step size of 0.05.
	bw : number or vector
		Numeric, specifies the bandwidth used for estimation. Can be (1) a positive scalar (common bandwidth for all grid points); or (2) a positive numeric vector specifying bandwidths for each grid point (should be the same length as *grid*).
	p : int
		Nonnegative integer, specifies the order of the local polynomial used to construct point estimates. (Default is *2*.)
	q : int
		Nonnegative integer, specifies the order of the local polynomial used to construct confidence intervals/bands (a.k.a. the bias correction order). Default is *p+1*. When set to be the same as *p*, no bias correction will be performed. Otherwise it should be strictly larger than *p*.
	v : int
		Nonnegative integer, specifies the derivative of the distribution function to be estimated. *0* for the distribution function, *1* (default) for the density funtion, etc.
	kernel : string
		Specifies the kernel function, should be one of *"triangular"*, *"uniform"*, and *"epanechnikov"*.
	scale : number
		Numeric, specifies how estimates are scaled. For example, setting this parameter to 0.5 will scale down both the point estimates and standard errors by half. Default is *1*. This parameter is useful if only part of the sample is employed for estimation, and should not be confused with *Cweights* or *Pweights*.
	massPoints : boolean
		*True* (default) or *False*, specifies whether point estimates and standard errors should be adjusted if there are mass points in the data.
	bwselect : string
		String, specifies the method for data-driven bandwidth selection. This option will be ignored if *bw* is provided. Options are (1) *"mse-dpi"* (default, mean squared error-optimal bandwidth selected for each grid point); (2) *"imse-dpi"* (integrated MSE-optimal bandwidth, common for all grid points); (3) *"mse-rot"* (rule-of-thumb bandwidth with Gaussian reference model); and (4) *"imse-rot"* (integrated rule-of-thumb bandwidth with Gaussian reference model).
	stdVar : boolean
		*True* (default) or *False*, specifies whether the data should be standardized for bandwidth selection.
	regularize : boolean
		*True* (default) or *False*, specifies whether the bandwidth should be regularized. When set to *True*, the bandwidth is chosen such that the local region includes at least *nLocalMin* observations and at least *nUniqueMin* unique observations.
	nLocalMin : int
		Nonnegative integer, specifies the minimum number of observations in each local neighborhood. This option will be ignored if *regularize=False*. Default is *20+p+1*.
	nUniqueMin : int
		Nonnegative integer, specifies the minimum number of unique observations in each local neighborhood. This option will be ignored if *regularize=False*. Default is *20+p+1*.
	Cweights : numeric vector
		Numeric, specifies the weights used for counterfactual distribution construction. Should have the same length as the data.
	Pweights : numeric vector
		Numeric, specifies the weights used in sampling. Should have the same length as the data.
	
	Returns
	-------
	Estimate:
		A matrix containing (1) *grid* (grid points), (2) *bw* (bandwidths), (3) *nh* (number of observations in each local neighborhood), (4) *nhu* (number of unique observations in each local neighborhood), (5) *f_p* (point estimates with p-th order local polynomial), (6) *f_q* (point estimates with q-th order local polynomial, only if option *q* is nonzero), (7) *se_p* (standard error corresponding to *f_p*), and (8) *se_q* (standard error corresponding to *f_q*).
	CovMat_p :
		The variance-covariance matrix corresponding to *f_p*.
	CovMat_q :
		The variance-covariance matrix corresponding to *f_q*.
	"""
	
	##############################
	##   INPUT ERROR HANDLING   ##
	##############################
	
	#data
	data = pd.DataFrame(data)
	if data.isnull().values.any():
		raise Exception(data.isnull().sum() + 'missing observation(s) are ignored.\n')
		data = data.dropna(axis=0)
	
	n = len(data)
	data = np.array(data).reshape(n,)
	if n==0 or not np.isscalar(n):
		raise Exception("Data should be numeric, and cannot be empty.")
	
	#grid
	if grid is None:
		grid = np.quantile(data, np.arange(0.05, 0.95, 0.05))
		ng = len(grid)
	else:
		ng = len(grid)
		if not np.isreal(grid).any():
			raise Exception("Grid points should be numeric.")

	if p is None:
		p = 2
	elif not np.isscalar(p) or p not in range(21):
		raise Exception("Polynomial order p incorrectly specified.")

	if q is None:
		q = p + 1
	elif not np.isscalar(q) or q not in range(21) or q<p:
		raise Exception("Polynomial order (for bias correction) q incorrectly specified")

	if v is None:
		v = min(1, p)
	elif not np.isscalar(q) or v not in range(21) or v>p:
		raise Exception("Derivative order v incorrectly specified")

	bw_list = ['mse-dpi','imse-dpi', 'mse-rot', 'imse-rot']
	if bw is None:
		if len(bwselect) == 0:
			bwselect = 'mse-dpi'
		else:
			bwselect = bwselect.lower()
			if bwselect not in bw_list:
				raise Exception("bwselect incorrectly specified")
	elif len(bw) == 1:
		if not np.isreal(bw) or any(bw) < 0:
			raise Exception("Bandwidth incorrectly specified.")
		else:
			bw = np.repeat(bw, ng)
			bwselect = 'user provided'
	else:
		bw = np.asarray(bw)
		if not np.isreal(bw):
			raise Exception("Bandwidth incorrectly specified.")
		elif len(bw) != ng:
			raise Exception("Bandwidth has to be the same length as grid.")
		else:
			bwselect = 'user provided'
			
	
	#kernel
	if kernel is None:
		flag_no_kernel = True
		kernel = 'epanechnikov'
	else:
		kernel = kernel.lower()
		if kernel not in ['triangular', 'uniform', 'epanechnikov']:
			raise Exception("Kernel function incorrectly specified.")

	# Cweights
	if Cweights is None:
		Cweights = np.repeat(1,n)
	elif not np.isreal(Cweights).any():
		raise Exception("Counterfactual weights incorrectly specified.")
	elif len(Cweights) !=n:
		raise Exception("Counterfactual weights should have the same length as sample.")
	
	if isinstance(Cweights, pd.Series):
		Cweights = Cweights.to_numpy()
	
	# Pweights
	if Pweights is None:
		Pweights = np.repeat(1,n)
	elif not np.isreal(Pweights).any():
		raise Exception("Probability weights incorrectly specified.")
	elif len(Pweights) !=n:
		raise Exception("Probability weights should have the same length as sample.")
	elif any(x<0 for x in Pweights)==True:
		raise Exception("Probability weights should be nonnegative.")
	
	if isinstance(Pweights, pd.Series):
		Pweights = Pweights.to_numpy()
	
	# scale
	if scale is None:
		scale = 1
	elif len(scale) != 1 or not np.isreal(scale) or scale[0] <= 0:
		raise Exception("Scale incorrectly specified.")
	else:
		scale = scale[0]

	# regularize
	if regularize is None:
		regularize = True
	elif type(regularize) != bool:
		raise Exception("Regularization parameter incorrectly specified")

	# massPoints
	if massPoints is None:
		massPoints = True
	elif type(massPoints) != bool:
		raise Exception("Option massPoints incorrectly specified.")


	# stdVar
	if stdVar is None:
		stdVar = True
	elif type(stdVar) != bool:
		raise Exception("Option stdVar incorrectly specified.")

	# nLocalMin
	if nLocalMin is None:
		nLocalMin = 20 + p + 1
	if not np.isreal(nLocalMin) or np.isnan(nLocalMin):
		raise Exception("Option nLocalMin incorrectly specified.")
	elif math.ceil(nLocalMin)<0:
		raise Exception("Option nLocalMin incorrectly specified.")
	else:
		nLocalMin = math.ceil(nLocalMin)


	# nUniqueMin
	if nUniqueMin is None:
		nUniqueMin = 20 + p + 1
	if not np.isreal(nUniqueMin) or np.isnan(nUniqueMin):
		raise Exception("Option nUniqueMin incorrectly specified.")
	elif math.ceil(nUniqueMin)<0:
		raise Exception("Option nUniqueMin incorrectly specified.")
	else:
		nUniqueMin = math.ceil(nUniqueMin)

	#########################
	##   Sample Trimming   ##
	#########################
	
	#	trim_index = (Pweights == 0)
#	if np.all(trim_index==True):
#		raise Exception("All weights are zero.")
#	else:
#		data = data[trim_index == False]
#		Cweights = Cweights[trim_index == False]
#		Pweights = Pweights[trim_index == False]

		if abs(sum(Cweights * Pweights)) <=np.finfo(float).eps * 10:
			raise Exception("Composited weights (Cweights * Pweights) are numerically zero.")

	#############################
	##   Bandwidth Selection   ##
	#############################
	if bwselect == 'mse-dpi':
		bw = lpbwdensity_fn.__bw_MSE( data=data, grid=grid, p=p, v=v, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=massPoints, stdVar=stdVar, regularize=regularize, nLocalMin=nLocalMin, nUniqueMin=nUniqueMin)
	elif bwselect == 'imse-dpi':
		bw =  lpbwdensity_fn.__bw_IMSE(data=data, grid=grid, p=p, v=v, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=massPoints, stdVar=stdVar, regularize=regularize, nLocalMin=nLocalMin, nUniqueMin=nUniqueMin)
		bw = np.repeat(bw, len(grid))
	elif bwselect == 'mse-rot':
		bw = lpbwdensity_fn.__bw_ROT( data=data, grid=grid, p=p, v=v, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=massPoints, stdVar=stdVar, regularize=regularize, nLocalMin=nLocalMin, nUniqueMin=nUniqueMin)
	else:
		bw = lpbwdensity_fn.__bw_IROT(data=data, grid=grid, p=p, v=v, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=massPoints, stdVar=stdVar, regularize=regularize, nLocalMin=nLocalMin, nUniqueMin=nUniqueMin)
		bw = np.repeat(bw, len(grid))
		
	##########################
	##   Point Estimation   ##
	##########################
	Temp_Result = lpdensity_fn.__lpdensity_fn(data=data, grid=grid, bw=bw, p=p, q=q, v=v, kernel=kernel,Cweights=Cweights, Pweights=Pweights, massPoints=massPoints, showSE=True)
	
	#scaling output
	Temp_Result['Estimate'][['f_p', 'f_q', 'se_p', 'se_q']] = Temp_Result['Estimate'][['f_p', 'f_q', 'se_p', 'se_q']] * scale
	Temp_Result['CovMat_p'] = Temp_Result['CovMat_p']* np.power(scale, 2)
	Temp_Result['CovMat_q'] = Temp_Result['CovMat_q']* np.power(scale, 2)

	
	#################################
	##   Return lpdensity Object   ##
	#################################
	
	lpoutput = lpdensity_output(Temp_Result['Estimate'], Temp_Result['CovMat_p'], Temp_Result['CovMat_q'], p, q, v, kernel, scale, massPoints, n, ng, bwselect, stdVar, regularize, nLocalMin, nUniqueMin, data_min = min(data), data_max = max(data), grid_min = min(grid), grid_max = max(grid))
	
	return(lpoutput)



#class of lpdensity output
class lpdensity_output:
	"""
	Class of lpdensity function outputs.
	
	Object type returned by :py:meth:`~lpdensity`.
	"""
	def __init__(self, Estimate, CovMat_p, CovMat_q, p, q, v, kernel, scale, massPoints, n, ng, bwselect, stdVar, regularize, nLocalMin, nUniqueMin, data_min, data_max, grid_min, grid_max):
		self.Estimate = Estimate
		self.CovMat_p = CovMat_p
		self.CovMat_q = CovMat_q
		self.p = p
		self.q = q
		self.v = v
		self.kernel = kernel
		self.scale = scale
		self.massPoints = massPoints
		self.n = n
		self.ng = ng
		self.bwselect = bwselect
		self.stdVar = stdVar
		self.regularize = regularize
		self.nLocalMin = nLocalMin
		self.nUniqueMin = nUniqueMin
		self.data_min = data_min
		self.data_max = data_max
		self.grid_min = grid_min
		self.grid_max = grid_max
	
	#summary output
	def __repr__(self):
		print('Call: lpdensity')
		print('')
		fw = 30
		fw_r = 14
		print('Sample Size:'.ljust(fw), str(self.n).rjust(25))
		print('Polynomial order for point estimation    (p=):'.ljust(fw), str(self.p).rjust(9))
		print('Order of derivative estimated            (v=):'.ljust(fw), str(self.v).rjust(9))
		print('Polynomial order for confidence interval (q=):'.ljust(fw), str(self.q).rjust(9))
		print('Kernel function:'.ljust(fw), str(self.kernel).rjust(25))
		print('Scaling factor:'.ljust(fw), str(self.scale).rjust(25))
		print('Bandwidth Method:'.ljust(fw), str(self.bwselect).rjust(25))
		print('')
		
		CIuniform=False
		alpha=0.05
		#critical value
		ci_df = pd.DataFrame(columns=['CI_l_q', 'CI_r_q'])

		if CIuniform==True:
			if np.isnan(CIsimul):
				raise warnings.warn('Option CIsimul incorrectly specified. Will only plot pointwise confidence intervals.')
				z_val = spstat.norm.ppf(1-alpha/2)
			elif math.ceil(CIsimul)<2:
				raise warnings.warn('Option CIsimul incorrectly specified. Will only plot pointwise confidence intervals.')
				z_val = spstat.norm.ppf(1-alpha/2)
			else:
				corrMat = self.CovMat_q.mul(1/self.Estimate['se_q'], axis=0).mul(1/self.Estimate['se_q'], axis=1)
				#cehck PSD matrix
				if np.all(np.linalg.eigvals(corrMat)>=0):
					#simulate multivariate normal
					normalSimul = np.random.multivariate_normal(mean = np.repeat(0, len(self.CovMat_q)), cov = corrMat, size=2000)
					#comput estimated quantile
					z_val = np.quantile(list(map(lambda x: max(abs(x)), normalSimul)), 1-alpha)
					
				else:
					#exception for non-PSD covariance matrix
					raise warnings.warn('Variance-Covariance is not positive semidefinite. Will only plot pointwise confidence intervals.')
					z_val = spstat.norm.ppf(1-alpha/2)
		else:
			z_val = spstat.norm.ppf(1-alpha/2)
		
		ci_df['CI_l_q'] = self.Estimate['f_q'] - z_val * self.Estimate['se_q']
		ci_df['CI_r_q'] = self.Estimate['f_q'] + z_val * self.Estimate['se_q']
		
		#Table
		fw_l = 15
		fw_c = 8
		fw_ci = 18
		n_dec = 3
		print('='*81)
		print('Index'.ljust(fw_c), 'Grid'.ljust(fw_c),
			 'B.W.'.ljust(fw_c), 'Eff.n'.ljust(fw_c), 'Uniq.n'.ljust(fw_c), 'Point.Est.'.rjust(fw_c),  'Std.Err'.rjust(fw_c), 'C.I.'.rjust(fw_c))
		print('-'*81)
		for j in range(self.ng):
			print(str(j+1).ljust(fw_c), str(round(self.Estimate['grid'][j], n_dec)).ljust(fw_c),
			str(round(self.Estimate['bw'][j], n_dec)).ljust(fw_c), str(int(self.Estimate['nh'][j])).ljust(fw_c), str(int(self.Estimate['nhu'][j])).ljust(fw_c), str(round(self.Estimate['f_p'][j], n_dec)).rjust(fw_c), str(round(self.Estimate['se_p'][j], n_dec)).rjust(fw_c), str('['+str(round(ci_df['CI_l_q'][j], n_dec)) + ',' + str(round(ci_df['CI_r_q'][j], n_dec))+']').rjust(fw_l))
		print('='*81)
		return('')
		
	#print call
	def __str__(self):
		print('Call: lpdensity')
		print('')
		fw = 30
		fw_r = 14
		print('Sample Size:'.ljust(fw), str(self.n).rjust(25))
		print('Polynomial order for point estimation    (p=):'.ljust(fw), str(self.p).rjust(9))
		print('Order of derivative estimated            (v=):'.ljust(fw), str(self.v).rjust(9))
		print('Polynomial order for confidence interval (q=):'.ljust(fw), str(self.q).rjust(9))
		print('Kernel function:'.ljust(fw), str(self.kernel).rjust(25))
		print('Scaling factor:'.ljust(fw), str(self.scale).rjust(25))
		print('Bandwidth Method:'.ljust(fw), str(self.bwselect).rjust(25))
		print('')
		print('Use .repr() to show estimates.')
		return ('')
	
	#coefficients of estimation
	def coef(self):
		"""
		Returns estimate coefficients.
		"""
		coef_data = self.Estimate[['grid', 'f_p', 'f_q']]
		return(coef_data)
	
	#variance-covariance
	def vcov(self):
		"""
		Returns estimate standard error and covariance matrices.
		"""
		vcov_data = self.Estimate[['grid', 'se_p', 'se_q']]
		return({'stdErr': vcov_data, 'CovMat_p': self.CovMat_p, 'CovMat_q': self.CovMat_q})
	
	#confidence interval function
	def confint(self, alpha=0.05, CIuniform=False, CIsimul=2000):
		"""
		Returns confindence intervals/bands for prespecified confidence level.
		
		alpha : number
			Confindence level, must be between 0 and 1.
		CIuniform : boolean
			Boolean on wehther to construct uniform confidence bands, *True* or *False (default)*.
		CIsimul : int
			Number of simulations used to generate confidence intervals.
		"""
		if alpha<0 or alpha>1:
			raise Exception("Significance level incorrectly specified.")
		
		#initialized output DF
		estimate_df = pd.DataFrame(columns=['grid', 'f_p', 'CI_l_p', 'CI_r_p', 'f_q', 'CI_l_q', 'CI_r_q'])
		
		estimate_df['grid'] = self.Estimate['grid']
		estimate_df['f_p'] = self.Estimate['f_p']
		estimate_df['f_q'] = self.Estimate['f_q']
		
		#critical value
		if CIuniform==True:
			if np.isnan(CIsimul):
				raise warnings.warn('Option CIsimul incorrectly specified. Will only plot pointwise confidence intervals.')
				z_val = spstat.norm.ppf(1-alpha/2)
			elif math.ceil(CIsimul)<2:
				raise warnings.warn('Option CIsimul incorrectly specified. Will only plot pointwise confidence intervals.')
				z_val = spstat.norm.ppf(1-alpha/2)
			else:
				CIsimul = math.ceil(CIsimul)
				corrMat = self.CovMat_q.mul(1/self.Estimate['se_q'], axis=0).mul(1/self.Estimate['se_q'], axis=1)
				#cehck PSD matrix
				if np.all(np.linalg.eigvals(corrMat)>=0):
					#simulate multivariate normal
					normalSimul = np.random.multivariate_normal(mean = np.repeat(0, len(self.CovMat_q)), cov = corrMat, size=CIsimul)
					#comput estimated quantile
					z_val = np.quantile(list(map(lambda x: max(abs(x)), normalSimul)), 1-alpha)
					
				else:
					#exception for non-PSD covariance matrix
					raise warnings.warn('Variance-Covariance is not positive semidefinite. Will only plot pointwise confidence intervals.')
					z_val = spstat.norm.ppf(1-alpha/2)
		else:
			z_val = spstat.norm.ppf(1-alpha/2)
		
		estimate_df['CI_l_p'] = self.Estimate['f_p'] - z_val * self.Estimate['se_p']
		estimate_df['CI_r_p'] = self.Estimate['f_p'] + z_val * self.Estimate['se_p']
		
		estimate_df['CI_l_q'] = self.Estimate['f_q'] - z_val * self.Estimate['se_q']
		estimate_df['CI_r_q'] = self.Estimate['f_q'] + z_val * self.Estimate['se_q']
		
		return(estimate_df)
	
	#plotting function
	def plot(self, alpha=0.05, type='line', CItype='region', CIuniform=False, CIsimul=2000, hist=False, histData=None, histBins=None, histFillCol=3, histFillShade=0.2, histLineCol="white", title=None, xlabel=None, ylabel=None, CIshade=0.2):
	
		"""
		Method to plot estimate and confidence bands.
		Requires ggplot.
		
		alpha : number
			Confindence level, must be between 0 and 1.
		CIuniform : boolean
			Boolean on wehther to construct uniform confidence bands, *True* or *False (default)*.
		CIsimul : int
			Number of simulations used to generate confidence intervals.
		type : string
			type of estimate plot, *line (default)*, or *points*, or *all*.
		CItype : string
			type of confidence interval plot, *region (default)*, *lines*, *ebar*, or *all*.
		"""
		if self.ng <2:
			raise Exception("At least two grid points are needed to plot input.")
			
		if alpha<0 or alpha>1:
			raise Exception("Significance level incorrectly specified.")
		#TODO: error handling
		if hist==True and histData is not None:
			histData = pd.DataFrame(histData, columns='d1')
			if histBins is None:
				histBreaks = 25
				temp_plot = (pn.ggplot(histData, pn.aes(x='d1')) + pn.geom_histogram(bins=histBins) + pn.theme_bw())
		else:
			temp_plot = pn.ggplot() + pn.theme_bw()
		
		
		data_x = self.Estimate[['grid', 'f_p', 'f_q', 'se_p', 'se_q']]
		
		#critical value
		if CIuniform==True:
			if np.isnan(CIsimul):
				raise warnings.warn('Option CIsimul incorrectly specified. Will only plot pointwise confidence intervals.')
				z_val = spstat.norm.ppf(1-alpha/2)
			elif math.ceil(CIsimul)<2:
				raise warnings.warn('Option CIsimul incorrectly specified. Will only plot pointwise confidence intervals.')
				z_val = spstat.norm.ppf(1-alpha/2)
			else:
				CIsimul = math.ceil(CIsimul)
				corrMat = self.CovMat_q.mul(1/self.Estimate['se_q'], axis=0).mul(1/self.Estimate['se_q'], axis=1)
				#cehck PSD matrix
				if np.all(np.linalg.eigvals(corrMat)>=0):
					#simulate multivariate normal
					normalSimul = np.random.multivariate_normal(mean = np.repeat(0, len(self.CovMat_q)), cov = corrMat, size=CIsimul)
					#comput estimated quantile
					z_val = np.quantile(list(map(lambda x: max(abs(x)), normalSimul)), 1-alpha)
					
				else:
					#exception for non-PSD covariance matrix
					raise warnings.warn('Variance-Covariance is not positive semidefinite. Will only plot pointwise confidence intervals.')
					z_val = spstat.norm.ppf(1-alpha/2)
		else:
			z_val = spstat.norm.ppf(1-alpha/2)
				
		data_x['CI_l'] = data_x['f_q'] - z_val * data_x['se_q']
		data_x['CI_r'] = data_x['f_q'] + z_val * data_x['se_q']
		
		if CItype in ['region', 'all']:
			temp_plot = temp_plot + pn.geom_ribbon(data=data_x, mapping=pn.aes(x='grid', ymin='CI_l', ymax='CI_r'), alpha=CIshade)
		
		if CItype in ['lines', 'all']:
			temp_plot = temp_plot + pn.geom_line(data=data_x, mapping=pn.aes(x='grid', y='CI_l')) + pn.geom_line(data=data_x, mapping=pn.aes(x='grid', y='CI_r'))
			
		if CItype in ['ebar', 'all']:
			temp_plot = temp_plot + pn.geom_errorbar(data=data_x, mapping=pn.aes(x='grid', ymin='CI_l', ymax='CI_r'))
		
		if type in ['line', 'both']:
			temp_plot = temp_plot + pn.geom_line(data=data_x, mapping=pn.aes(x='grid', y='f_p'))
		
		if type in ['points', 'both']:
			temp_plot = temp_plot + pn.geom_point(data=data_x, mapping=pn.aes(x='grid', y='f_p'))
		
		if xlabel is None:
			xlabel = ''
		
		if ylabel is None:
			ylabel = ''
		
		if title is None:
			title = ''
			
		temp_plot = temp_plot + pn.xlab(xlabel) + pn.ylab(ylabel)
		
		return(temp_plot)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#ignore future warnings from pandas
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)

import numpy as np
import pandas as pd
import math
from numpy.linalg import inv
from . import lpbwdensity_fn

def __lpdensity_fn(data, grid, bw, p, q, v, kernel, Cweights, Pweights, massPoints, showSE=True):
	# data prep
	ii = np.argsort(data)
	data = data[ii]
	Cweights = Cweights[ii]
	Pweights = Pweights[ii]
	n = len(data)
	ng = len(grid)
	
	dataUnique = lpbwdensity_fn.__lpdensityUnique(data)
	freqUnique = dataUnique["freq"]
	indexUnique = dataUnique["index"]
	dataUnique = dataUnique["unique"]
	nUnique = len(dataUnique)

	# whether considering mass points when constructing the empirical distribution function
	if massPoints == True:
		Fn = np.repeat((np.cumsum(Cweights * Pweights) / sum(Cweights * Pweights))[indexUnique], repeats = freqUnique)
	else:
		Fn = np.cumsum(Cweights * Pweights) / sum(Cweights * Pweights)
	
	weights_normal = Cweights * Pweights / sum(Cweights * Pweights) * n
	Cweights = Cweights / sum(Cweights) * n
	Pweights = Pweights / sum(Pweights) * n
	
	weights_normalUnique = np.cumsum(weights_normal)[indexUnique]
	if nUnique > 1:
		weights_normalUnique = weights_normalUnique - np.insert(weights_normalUnique[0:(nUnique-1)],0,0)
	
	CweightsUnique = np.cumsum(Cweights)[indexUnique]
	if nUnique > 1:
		CweightsUnique = CweightsUnique - np.insert(CweightsUnique[0:(nUnique-1)],0,0)
	
	PweightsUnique = np.cumsum(Pweights)[indexUnique]
	if nUnique > 1:
		PweightsUnique = PweightsUnique - np.insert(PweightsUnique[0:(nUnique-1)],0,0)
	
	# initialization
	hat_p = np.repeat(float("NaN"), ng)
	hat_q = np.repeat(float("NaN"), ng)
	se_p = np.repeat(float("NaN"), ng)
	se_q = np.repeat(float("NaN"), ng)
	iff_p = np.empty((n, ng))
	iff_p[:] = np.NaN
	iff_q = np.empty((n, ng))
	iff_q[:] = np.NaN
	nh = np.repeat(float("NaN"), ng)
	nhu = np.repeat(float("NaN"), ng)
	
	# looping over grid points
	for j in range(ng) :
		index_temp = abs(data-grid[j]) <= bw[j]
		nh[j] = sum(index_temp)
		nhu[j] = sum(index_temp[indexUnique])
			
		# LHS and RHS variables
		if massPoints == True:
			Y_temp = pd.DataFrame(Fn[indexUnique])
			Xh_temp = pd.DataFrame(data[indexUnique]-grid[j])/bw[j]
			Xh_p_temp = Xh_temp[np.repeat(Xh_temp.columns, p+1)]
			Xh_p_temp = Xh_p_temp.apply(lambda x: np.power(x, range(p+1)), axis=1)
			
			# weights
			if kernel=="triangular":
				Kh_temp = (1-abs(Xh_temp))/bw[j]
				Kh_temp = Kh_temp.mul(index_temp[indexUnique], axis=0)
			elif kernel=="uniform":
				Kh_temp = (0.5/bw[j]) * index_temp[indexUnique]
			else :
				Kh_temp = 0.75*(1-np.power(Xh_temp,2))/bw[j]
				Kh_temp = Kh_temp.mul(index_temp[indexUnique], axis=0)
			
			Xh_p_Kh_temp = Xh_p_temp.mul(Kh_temp, axis=0)
			Xh_p_Kh_Pweights_temp = Xh_p_Kh_temp.mul(PweightsUnique, axis=0)
			
			try:
				XhKhXh_inv = inv(np.matmul(Xh_p_temp[index_temp[indexUnique]].transpose(), Xh_p_Kh_Pweights_temp[index_temp[indexUnique]])/ n)
			except np.linalg.LinAlgError:
				continue
			
			#point estimate
			hat_p[j] = math.factorial(v) * np.matmul(Y_temp.transpose(), np.matmul(Xh_p_Kh_Pweights_temp, XhKhXh_inv))[v]/(np.power(bw[j],v)) /n
		else:
			Y_temp = pd.DataFrame(Fn)
			Xh_temp = pd.DataFrame(data-grid[j])/bw[j]
			Xh_p_temp = Xh_temp[np.repeat(Xh_temp.columns, p+1)]
			Xh_p_temp = Xh_p_temp.apply(lambda x: np.power(x, range(p+1)), axis=1)
			
			# weights
			if kernel=="triangular":
				Kh_temp = ((1-abs(Xh_temp))/bw[j]) * index_temp
			elif kernel=="uniform":
				Kh_temp = (0.5/bw[j]) * index_temp
			else :
				Kh_temp = (0.75*(1-np.power(Xh_temp, 2))/bw[j]) * index_temp
				
			Xh_p_Kh_temp = Xh_p_temp.mul(Kh_temp, axis=0)
			Xh_p_Kh_Pweights_temp = Xh_p_temp.mul(PweightsUnique, axis=0)
			
			try:
				XhKhXh_inv = inv(np.matmul(Xh_p_temp[index_temp].transpose(), Xh_p_Kh_Pweights_temp[index_temp])/ n)
			except np.linalg.LinAlgError:
				continue
			
			#point estimate
			hat_p[j] = math.factorial(v) * np.matmul(XhKhXh_inv, np.matmul(Xh_p_Kh_Pweights_temp[index_temp,:].transpose(), Y_temp[index_temp]))[v]/(np.power(bw[j], v)) /n
			
		if showSE==True:
			if massPoints==True:
				F_Xh_p_Kh_temp = np.matmul(Y_temp.transpose(), Xh_p_Kh_Pweights_temp)/n
				
				G = np.zeros((n,  len(Xh_p_Kh_Pweights_temp.columns)))
				for jj in range(len(Xh_p_Kh_Pweights_temp.columns)):
					G[:,jj] = (np.repeat((np.cumsum(Xh_p_Kh_Pweights_temp.iloc[::-1, jj])/n)[::-1], freqUnique) - F_Xh_p_Kh_temp.loc[0, jj]) * weights_normal
				
				G = pd.DataFrame(G)
				
				iff_p[:,j] = np.matmul(XhKhXh_inv, G.transpose()).iloc[v, :] * math.factorial(v)/np.sqrt(n*np.power(bw[j], 2*v))
			else:
				F_Xh_p_Kh_temp = np.matmul(Y_temp.transpose(), Xh_p_Kh_Pweights_temp)/n
				
				G = np.zeros((n,  len(Xh_p_Kh_Pweights_temp.columns)))
				for jj in range(len(Xh_p_Kh_Pweights_temp.columns)):
					G[:,jj] = (np.repeat(np.cumsum(Xh_p_Kh_Pweights_temp.iloc[:, jj]/n), freqUnique) - F_Xh_p_Kh_temp.loc[0, jj]) * weights_normal
				
				G = pd.DataFrame(G)
				
				iff_p[:,j] = np.matmul(XhKhXh_inv, G.transpose()).iloc[v, :] * math.factorial(v)/np.sqrt(n*np.power(bw[j], 2*v))
				
		if q>p:
			if massPoints == True:
				Y_temp = pd.DataFrame(Fn[indexUnique])
				Xh_temp = pd.DataFrame(data[indexUnique]-grid[j])/bw[j]
				Xh_q_temp = Xh_temp[np.repeat(Xh_temp.columns, q+1)]
				Xh_q_temp = Xh_q_temp.apply(lambda x: np.power(x, range(q+1)), axis=1)
			
				# weights
				if kernel=="triangular":
					Kh_temp = (1-abs(Xh_temp))/bw[j]
					Kh_temp = Kh_temp.mul(index_temp[indexUnique], axis=0)
				elif kernel=="uniform":
					Kh_temp = (0.5/bw[j]) * index_temp[indexUnique]
				else :
					Kh_temp = 0.75*(1-np.power(Xh_temp,2))/bw[j]
					Kh_temp = Kh_temp.mul(index_temp[indexUnique], axis=0)
				
				Xh_q_Kh_temp = Xh_q_temp.mul(Kh_temp, axis=0)
				Xh_q_Kh_Pweights_temp = Xh_q_Kh_temp.mul(PweightsUnique, axis=0)
				
				try:
					XhKhXh_inv = inv(np.matmul(Xh_q_temp[index_temp[indexUnique]].transpose(), Xh_q_Kh_Pweights_temp[index_temp[indexUnique]])/ n)
				except np.linalg.LinAlgError:
					continue
				
				#point estimate
				hat_q[j] = math.factorial(v) * np.matmul(Y_temp.transpose(), np.matmul(Xh_q_Kh_Pweights_temp, XhKhXh_inv))[v]/(np.power(bw[j],v)) /n
			else:
				Y_temp = pd.DataFrame(Fn)
				Xh_temp = pd.DataFrame(data-grid[j])/bw[j]
				Xh_q_temp = Xh_temp[np.repeat(Xh_temp.columns, q+1)]
				Xh_q_temp = Xh_q_temp.apply(lambda x: np.power(x, range(q+1)), axis=1)
				
				# weights
				if kernel=="triangular":
					Kh_temp = ((1-abs(Xh_temp))/bw[j]) * index_temp
				elif kernel=="uniform":
					Kh_temp = (0.5/bw[j]) * index_temp
				else :
					Kh_temp = (0.75*(1-np.power(Xh_temp, 2))/bw[j]) * index_temp
					
				Xh_q_Kh_temp = Xh_q_temp.mul(Kh_temp, axis=0)
				Xh_q_Kh_Pweights_temp = Xh_q_temp.mul(PweightsUnique, axis=0)
				
				try:
					XhKhXh_inv = inv(np.matmul(Xh_q_temp[index_temp].transpose(), Xh_q_Kh_Pweights_temp[index_temp])/ n)
				except np.linalg.LinAlgError:
					continue
				
				#point estimate
				hat_q[j] = math.factorial(v) * np.matmul(XhKhXh_inv, np.matmul(Xh_q_Kh_Pweights_temp[index_temp,:].transpose(), Y_temp[index_temp]))[v]/(np.power(bw[j], v)) /n
				
			if showSE==True:
				if massPoints==True:
					F_Xh_q_Kh_temp = np.matmul(Y_temp.transpose(), Xh_q_Kh_Pweights_temp)/n
					
					G = np.zeros((n,  len(Xh_q_Kh_Pweights_temp.columns)))
					for jj in range(len(Xh_q_Kh_Pweights_temp.columns)):
						G[:,jj] = (np.repeat((np.cumsum(Xh_q_Kh_Pweights_temp.iloc[::-1, jj])/n)[::-1], freqUnique) - F_Xh_q_Kh_temp.loc[0, jj]) * weights_normal
					
					G = pd.DataFrame(G)
					
					iff_q[:,j] = np.matmul(XhKhXh_inv, G.transpose()).iloc[v, :] * math.factorial(v)/np.sqrt(n*np.power(bw[j], 2*v))
				else:
					F_Xh_q_Kh_temp = np.matmul(Y_temp.transpose(), Xh_q_Kh_Pweights_temp)/n
					
					G = np.zeros((n,  len(Xh_q_Kh_Pweights_temp.columns)))
					for jj in range(len(Xh_q_Kh_Pweights_temp.columns)):
						G[:,jj] = (np.repeat(np.cumsum(Xh_q_Kh_Pweights_temp.iloc[:, jj]/n), freqUnique) - F_Xh_q_Kh_temp.loc[0, jj]) * weights_normal
					
					G = pd.DataFrame(G)
					
					iff_q[:,j] = np.matmul(XhKhXh_inv, G.transpose()).iloc[v, :] * math.factorial(v)/np.sqrt(n*np.power(bw[j], 2*v))
					

	# putting together all the data in the right dataframes
	
	CovMat_p = pd.DataFrame(data=np.matmul(iff_p.transpose(), iff_p)/n)
	CovMat_q = pd.DataFrame(data=np.matmul(iff_q.transpose(), iff_q)/n)

	se_p = np.sqrt(abs(np.diag(CovMat_p)))
	se_q = np.sqrt(abs(np.diag(CovMat_q)))
	
	Estimate = np.empty((ng, 8))
	Estimate[:,0] = grid
	Estimate[:,1] = bw
	Estimate[:,2] = nh
	Estimate[:,3] = nhu
	Estimate[:,4] = hat_p
	Estimate[:,5] = hat_q
	Estimate[:,6] = se_p
	Estimate[:,7] = se_q
	
	est_data = pd.DataFrame(data=Estimate, columns=["grid", "bw", "nh", "nhu", "f_p", "f_q", "se_p", "se_q"])
	
	return({'Estimate': est_data, 'CovMat_p': CovMat_p, 'CovMat_q': CovMat_q})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
import math
from . import lpbwdensity_fn

pd.options.mode.chained_assignment = None

def lpbwdensity(data, grid=None, p=None, v=None, kernel="triangular", bwselect="mse-dpi", massPoints=True, stdVar=True, regularize=True, nLocalMin=None, nUniqueMin=None, Cweights=None, Pweights=None):
	"""
	Parameters
	----------
		 data : vector
			Numeric vector or one dimensional matrix/data frame, the raw data.
		 grid : vector
			Numeric, specifies the grid of evaluation points. When set to default, grid points will be chosen as 0.05-0.95 percentiles of the data, with a step size of 0.05.
		 p : int
			Nonnegative integer, specifies the order of the local polynomial used to construct point estimates. (Default is *2*.)
		 v : int
			Nonnegative integer, specifies the derivative of the distribution function to be estimated. *0* for the distribution function, *1* (default) for the density funtion, etc.
		 kernel : string
			Specifies the kernel function, should be one of *"triangular"*, *"uniform"* or
		   *"epanechnikov"*.
		 bwselect : string
			Specifies the method for data-driven bandwidth selection. This option will be
			ignored if *bw* is provided. Can be (1) *"mse-dpi"* (default, mean squared error-optimal
			bandwidth selected for each grid point); or (2) *"imse-dpi"* (integrated MSE-optimal bandwidth,
			common for all grid points); (3) *"mse-rot"* (rule-of-thumb bandwidth with Gaussian
			reference model); and (4) *"imse-rot"* (integrated rule-of-thumb bandwidth with Gaussian
			reference model).
		 massPoints : boolean
			*True* (default) or *False*, specifies whether point estimates and standard errors
		   should be adjusted if there are mass points in the data.
		 stdVar : boolean
			*True* (default) or *False*, specifies whether the data should be standardized for
		   bandwidth selection.
		 regularize : boolean
			*True* (default) or *False*, specifies whether the bandwidth should be
		   regularized. When set to *True*, the bandwidth is chosen such that the local region includes
		   at least *nLocalMin* observations and at least *nUniqueMin* unique observations.
		 nLocalMin : int
			Nonnegative integer, specifies the minimum number of observations in each local neighborhood. This option
		   will be ignored if *regularize=False*. Default is *20+p+1*.
		 nUniqueMin : int
			Nonnegative integer, specifies the minimum number of unique observations in each local neighborhood. This option
		   will be ignored if *regularize=False*. Default is *20+p+1*.
		 Cweights : vector
			Numeric vector, specifies the weights used
		   for counterfactual distribution construction. Should have the same length as the data.
		   This option will be ignored if *bwselect* is *"mse-rot"* or *"imse-rot"*.
		 Pweights : vector
			Numeric vector, specifies the weights used
		   in sampling. Should have the same length as the data.
		   This option will be ignored if *bwselect* is *"mse-rot"* or *"imse-rot"*.
		   
	Returns
	-------
	BW : object
		A :py:meth:`~bw_output` class object containing a matrix of (1) *grid* (grid point), (2) *bw* (bandwidth),
		(3) *nh* (number of observations in each local neighborhood),
		(4) *nhu* (number of unique observations in each local neighborhood), and
		(5) *opt* additional parameters.
	"""

	# data
	data = pd.DataFrame(data)
	if data.isnull().values.any():
		raise Exception(data.isnull().sum() + 'missing observation(s) are ignored.\n')
		data = data.dropna(axis=0)
	
	n = len(data)
	data = np.array(data).reshape(n,)
	if n==0 or not np.isscalar(n):
		raise Exception('Data should be numeric, and cannot be empty.')
	
	# grid
	if grid is None:
		grid = np.quantile(data, np.arange(0.05, 0.95, 0.05))
		ng = len(grid)
	else:
		ng = len(grid)
		if not np.isreal(grid).any():
			raise Exception('Grid points should be numeric.')
	
	# bwselect
	if len(bwselect) == 0:
		bwselect = "mse-dpi"
	else:
		bwselect = bwselect.lower()
	
	# p
	if p is None:
		p = 2
	elif not np.isscalar(p) or p not in range(21):
		raise Exception('Polynomial order p incorrectly specified.')
		
	# v
	if v is None:
		v = min(1, p)
	elif not np.isscalar(v) or v not in range(21) or v>p:
		raise Exception('Derivative order v incorrectly specified')
		
	#kernel
	if kernel is None:
		flag_no_kernel = True
		kernel = "triangular"
	else:
		kernel = kernel.lower()
		if kernel not in ["triangular", "uniform", "epanechnikov"]:
			raise Exception('Kernel function incorrectly specified.')
	
	# Cweights
	if Cweights is None:
		Cweights = np.repeat(1,n)
	elif not np.isreal(Cweights).any():
		raise Exception('Counterfactual weights incorrectly specified.')
	elif len(Cweights) !=n:
		raise Exception('Counterfactual weights should have the same length as sample.')
	
	if isinstance(Cweights, pd.Series):
		Cweights = Cweights.to_numpy()
	
	# Pweights
	if Pweights is None:
		Pweights = np.repeat(1,n)
	elif not np.isreal(Pweights).any():
		raise Exception('Probability weights incorrectly specified.')
	elif len(Pweights) !=n:
		raise Exception('Probability weights should have the same length as sample.')
	elif any(x<0 for x in Pweights)==True:
		raise Exception('Probability weights should be nonnegative.')
		
	if isinstance(Pweights, pd.Series):
		Pweights = Pweights.to_numpy()
	
	# massPoints
	if massPoints is None:
		massPoints = True
	elif type(massPoints) != bool:
		raise Exception('Option massPoints incorrectly specified.')
	
	
	# stdVar
	if stdVar is None:
		stdVar = True
	elif type(stdVar) != bool:
		raise Exception('Option stdVar incorrectly specified.')
	
	
	# regularize
	if regularize is None:
		regularize = True
	elif type(regularize) != bool:
		raise Exception('Regularization parameter incorrectly specified')
	
	
	# nLocalMin
	if nLocalMin is None:
		nLocalMin = 20 + p + 1
	if not np.isreal(nLocalMin) or np.isnan(nLocalMin):
		raise Exception('Option nLocalMin incorrectly specified.')
	elif math.ceil(nLocalMin)<0:
		raise Exception('Option nLocalMin incorrectly specified.')
	else:
		nLocalMin = math.ceil(nLocalMin)

	# nUniqueMin
	if nUniqueMin is None:
		nUniqueMin = 20 + p + 1
	if not np.isreal(nUniqueMin) or np.isnan(nUniqueMin):
		raise Exception('Option nUniqueMin incorrectly specified.')
	elif math.ceil(nUniqueMin)<0:
		raise Exception('Option nUniqueMin incorrectly specified.')
	else:
		nUniqueMin = math.ceil(nUniqueMin)
		
	#########################
	##   Sample Trimming   ##
	#########################
#	trim_index = (Pweights==0)
#	if np.all(trim_index==True):
#		raise Exception('All weights are zero.')
#	else:
#		data = data[Pweights!=0]
#		Cweights = Cweights[Pweights!=0]
#		Pweights = Pweights[Pweights!=0]
	
	if abs(sum(Cweights*Pweights)) <= 10*np.finfo(float).eps:
		raise Exception('Composited weights (Cweights * Pweights) are numerically zero.')
		
	#############################
	##   Bandwidth Selection   ##
	#############################
	
	if bwselect == "mse-dpi":
		bw = lpbwdensity_fn.__bw_MSE( data=data, grid=grid, p=p, v=v, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=massPoints, stdVar=stdVar, regularize=regularize, nLocalMin=nLocalMin, nUniqueMin=nUniqueMin)
	elif bwselect == "imse-dpi":
		bw =  lpbwdensity_fn.__bw_IMSE(data=data, grid=grid, p=p, v=v, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=massPoints, stdVar=stdVar, regularize=regularize, nLocalMin=nLocalMin, nUniqueMin=nUniqueMin)
		bw = np.repeat(bw, len(grid))
	elif bwselect == "mse-rot":
		bw = lpbwdensity_fn.__bw_ROT( data=data, grid=grid, p=p, v=v, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=massPoints, stdVar=stdVar, regularize=regularize, nLocalMin=nLocalMin, nUniqueMin=nUniqueMin)
	else:
		bw = lpbwdensity_fn.__bw_IROT(data=data, grid=grid, p=p, v=v, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=massPoints, stdVar=stdVar, regularize=regularize, nLocalMin=nLocalMin, nUniqueMin=nUniqueMin)
		bw = np.repeat(bw, len(grid))
		
	BW = pd.DataFrame(columns = ["grid", "bw", "nh", "nhu"])
	BW['grid'] = grid
	BW['bw'] = bw
	
	dataUnique =  lpbwdensity_fn.__lpdensityUnique(np.sort(data))['unique']
	
	for i in range(ng):
		BW['nh'][i] = sum(abs(data-BW['grid'][i])<= BW['bw'][i])
		BW['nhu'][i] = sum(abs(dataUnique-BW['grid'][i]) <= BW['bw'][i])
	
	return(bw_output(BW, bw, p, v, kernel, n, ng, bwselect, massPoints, stdVar, regularize, nLocalMin, nUniqueMin, data_min=min(data), data_max=max(data), grid_min=min(grid), grid_max=max(grid)))


class bw_output:
	"""
	Class of lpbwdensity function outputs.
	
	Object type returned by :py:meth:`~lpbwdensity`.
	"""
	def __init__(self, BW, bws, p, v, kernel, n, ng, bwselect, massPoints, stdVar, regularize, nLocalMin, nUniqueMin, data_min, data_max, grid_min, grid_max):
		self.table = BW
		self.bws = bws
		self.p = p
		self.v = v
		self.kernel = kernel
		self.n = n
		self.ng = ng
		self.bwselect = bwselect
		self.massPoints = massPoints
		self.stdVar = stdVar
		self.regularize = regularize
		self.nLocalMin = nLocalMin
		self.nUniqueMin = nUniqueMin
		self.data_min = data_min
		self.data_max = data_max
		self.grid_min = grid_min
		self.grid_max = grid_max
	
	def __repr__(self):
		print('Call: lpbwdensity')
		print('')
		fw = 30
		fw_r = 14
		print('Sample Size:'.ljust(fw), str(self.n).rjust(25))
		print('Polynomial order for point estimation (p=):'.ljust(fw), str(self.p).rjust(12))
		print('Order of derivative estimated         (v=):'.ljust(fw), str(self.v).rjust(12))
		print('Kernel function:'.ljust(fw), str(self.kernel).rjust(25))
		print('Bandwidth Method:'.ljust(fw), str(self.bwselect).rjust(25))
		print('')
		
		#Table
		fw_l = 15
		fw_c = 8
		fw_ci = 18
		n_dec = 3
		print('='*65)
		print('Index'.ljust(fw_c), 'Grid'.rjust(fw_c),
			'Bandwidths'.rjust(fw_l),
			'Eff.n'.rjust(fw_c), 'Uniq.n'.rjust(fw_c))
		print('-'*65)
		for j in range(self.ng):
			print(str(j+1).ljust(fw_c), str(round(self.table['grid'][j], n_dec)).rjust(fw_c),
			str(round(self.table['bw'][j], n_dec)).rjust(fw_l), str(int(self.table['nh'][j])).rjust(fw_c), str(int(self.table['nhu'][j])).rjust(fw_c))
		print('='*65)
		return('')
		
	def __str__(self):
		print('Call: lpbwdensity')
		print('')
		fw = 30
		fw_r = 14
		print('Sample Size:'.ljust(fw), str(self.n).rjust(25))
		print('Polynomial order for point estimation (p=):'.ljust(fw), str(self.p).rjust(12))
		print('Order of derivative estimated         (v=):'.ljust(fw), str(self.v).rjust(12))
		print('Kernel function:'.ljust(fw), str(self.kernel).rjust(25))
		print('Bandwidth Method:'.ljust(fw), str(self.bwselect).rjust(25))
		print('')
		print('Use .repr() to show bandwidths.')
		return ('')
	
	def coef(self):
		"""
		Returns estimate of bandwidths.
		"""
		coef_data = self.table[['grid', 'bw']]
		return(coef_data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#ignore future warnings from pandas
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import numpy as np
from scipy.stats import norm
import sympy
from sympy.abc import x, y
import math
import scipy.integrate as integrate
from numpy.linalg import inv
import scipy.optimize as optimize
import statistics as stat
import pandas as pd


def __kernel_fn(a, h, kernel):
	if kernel == "triangular":
		kh = (1-abs(a))/h
	elif kernel == "uniform":
		kh = np.repeat(0.5/h, len(a))
	else:
		kh = 0.75*(1-(a)**2)/h
	return(kh)

def __lpdensityUnique(x):
	n = len(x)
	x = pd.DataFrame(x)
	# if x has less than 2 elements
	if n==0:
		return({'unique':None, 'freq':[None], 'index':[None]})
	if n==1:
		return({'unique':x, 'freq':[1], 'index':[0]})
		
	#if x has more than 1 element
	unique, uniqueIndex, nUniquelist = np.unique(x, return_index=True, return_counts=True, axis=0)
	nUnique = sum(unique)
#	uniqueIndex = x != x.shift()
#	unique = np.array(x)[uniqueIndex]
#	nUnique = len(unique)
	
	#if all are distinct
	if nUnique==n:
		return({'unique':unique, 'freq':np.repeat(1, len(x)), 'index':list(range(n))})
		
	#if all are the same
	if nUnique ==1:
		return({'unique':unique, 'freq':[n], 'index':[n-1]})
	
#	freq = np.cumsum(uniqueIndex==False)[uniqueIndex].iloc[:,0]
#	freq = freq - pd.concat([pd.Series([0]), freq[:-1]]) + 1
#	freq = pd.DataFrame(freq).reset_index(drop=True).dropna()[0]
	
	return({'unique':unique, 'freq':nUniquelist, 'index':uniqueIndex})

# Internal function
# Gaussian derivatives
# Output matches R
def __normal_dgps(x_val, v, mean, sd):
	if v==0:
		return(norm.cdf(x_val, loc=mean, scale=sd))
	else:
		
		temp = sympy.exp(-(x-mean)**2/(2*sd**2))/sympy.sqrt(2*math.pi*sd**2)
		while v>1:
			temp = sympy.diff(temp, x)
			v = v-1
		return(temp.evalf(subs={x:x_val}))

# Generate S matrix
# Output matches R
def __Sgenerate(p, low=-1, up=1, kernel="triangular"):
	S = np.zeros((p+1, p+1))
	for i in range(1,p+2):
		for j in range(1,p+2):
			if kernel=="uniform":
				integrand = lambda l: 0.5*np.power(l, i+j-2)
			elif kernel=="epanechnikov":
				integrand = lambda l: 0.75*np.power(l, i+j-2)*(1-l**2)
			else:
				integrand = lambda l: (1-abs(l))*np.power(l, i+j-2)
			
			S[i-1, j-1] = integrate.quad(integrand, low, up)[0]
	
	return(S)
	

# Generate T matrix
# Output matches R
def __Tgenerate(p, low=-1, up=1, kernel="triangular"):
	S = np.zeros((p+1, p+1))
	for i in range(1,p+2):
		for j in range(1,p+2):
			if kernel=="uniform":
				integrand = lambda l: np.power(0.5, 2)*np.power(l, i+j-2)
			elif kernel=="epanechnikov":
				integrand = lambda l: np.power(0.75*(1-l**2), 2)*np.power(l, i+j-2)
			else:
				integrand = lambda l: np.power(1-abs(l), 2)*np.power(l, i+j-2)
			
			S[i-1, j-1] = integrate.quad(integrand, low, up)[0]
	
	return(S)

# Generate C matrix
# Output matches R
def __Cgenerate(k, p, low=-1, up=1, kernel="triangular"):
	C = np.zeros((p+1, 1))
	for i in range(1, p+2):
		if kernel=="uniform":
			integrand = lambda l: 0.5*np.power(l, i+k-1)
		elif kernel=="epanechnikov":
			integrand = lambda l: 0.75*np.power(l, i+k-1)*(1-l**2)
		else:
			integrand = lambda l: (1-abs(l))*np.power(l, i+k-1)
			
		C[i-1, 0] = integrate.quad(integrand, low, up)[0]
	
	return(C)

# Generate G matrix
#TODO: Fix triangular kernel
def __Ggenerate(p, low=-1, up=1, kernel="uniform"):
	G = np.zeros((p+1, p+1))
	for i in range(1, p+2):
		for j in range(1, p+2):
			def integrand_1(x,y):
				return(np.power(x,i) * np.power(y, (j-1))*(1-abs(x))*(1-abs(y)))
				
			def integrand_2(x,y):
				return(np.power(x,i-1) * np.power(y, j)*(1-abs(x))*(1-abs(y)))
				
			def x_integral_1(y):
				return (integrate.quad(integrand_1, low, y, args=(y))[0])
			
			def x_integral_2(y):
				return (integrate.quad(integrand_2, y, up, args=(y))[0])
			
			G[i-1, j-1] = integrate.quad(x_integral_1, low, up)[0] + integrate.quad(x_integral_2, low, up)[0]
				
	return(G)



# MSE-ROT Bandwidth function
# matches R output
def __bw_ROT(data, grid, p, v, kernel, Cweights, Pweights, massPoints, stdVar, regularize, nLocalMin, nUniqueMin):
	n = len(data)
	ng = len(grid)
	
	dataUnique = np.unique(data)
	nUnique = len(dataUnique)
	
	if stdVar==True:
		center_temp = stat.mean(data)
		scale_temp = stat.stdev(data)
		data = (data-center_temp)/scale_temp
		dataUnique = (dataUnique-center_temp)/scale_temp
		grid = (grid-center_temp)/scale_temp
		
	#estimate normal reference model
	mean_hat = sum(Cweights*Pweights*data)/sum(Cweights*Pweights)
	sd_hat = np.sqrt(sum(Cweights*Pweights*(data-mean_hat)**2)/sum(Cweights*Pweights))
	
	#normal quantities
	temp_1 = sympy.exp(-(x-mean_hat)**2/(2*sd_hat**2))/sympy.sqrt(2*math.pi*sd_hat**2)
	temp_2 = sympy.diff(temp_1, x)
	temp_3 = sympy.exp(-(x-mean_hat)**2/(2*sd_hat**2))/sympy.sqrt(2*math.pi*sd_hat**2)
	temp_4 = sympy.exp(-(x-mean_hat)**2/(2*sd_hat**2))/sympy.sqrt(2*math.pi*sd_hat**2)
	
	j = p+1
	while j>1:
		temp_3 = sympy.diff(temp_3, x)
		j = j-1
	
	j=p+2
	while j >1:
		temp_4 = sympy.diff(temp_4, x)
		j = j-1
		
	#bias estimate, no rate added, DGP constant
	bias_dgp = np.zeros((ng, 2))
	for j in range(ng):
		bias_dgp[j, 0] = temp_3.evalf(subs={x:grid[j]})/math.factorial(p+1) * math.factorial(v)
		bias_dgp[j, 1] = temp_4.evalf(subs={x:grid[j]})/math.factorial(p+2) * math.factorial(v) + bias_dgp[j,0] * temp_2.evalf(subs={x:grid[j]})/temp_1.evalf(subs={x:grid[j]})
	
	#bias estimate, no rate added, kernel constant
	S = __Sgenerate(p=p, low=-1, up=1, kernel=kernel)
	C1 = __Cgenerate(k=p+1, p=p, low=-1, up=1, kernel=kernel)
	C2 = __Cgenerate(k=p+2, p=p, low=-1, up=1, kernel=kernel)
	G = __Ggenerate(p=p, low=-1, up=1, kernel=kernel)
	S2 = __Tgenerate(p=p, low=-1, up=1, kernel=kernel)
	bias_dgp[:,0] = bias_dgp[:,0] * (np.matmul(inv(S), C1)[v,:])
	bias_dgp[:,1] = bias_dgp[:,1] * (np.matmul(inv(S), C2)[v,:])
	
	#variance estimate, sample size added
	sd_dgp = np.zeros((ng, 1))
	if v>0:
		for j in range(ng):
			temp_1eval = np.array(temp_1.evalf(subs={x:grid[j]}), dtype=np.float64)
			sd_dgp[j, 0] = math.factorial(v) * np.sqrt(temp_1eval/n)
			
		sd_dgp = sd_dgp * np.sqrt(abs((inv(S) * G * inv(S))[v, v]))
		
	else:
		for j in range(ng):
			#this comes from a higher-order variance expansion. See Lemma 4 in the Appendix of Cattaneo, Jansson and Ma (2019a)
			sd_dgp[j, 0] = np.sqrt(norm.cdf(grid[j], loc=mean_hat, scale=sd_hat) * (1-norm.cdf(grid[j], loc=mean_hat, scale=sd_hat))/norm.pdf(grid[j], loc=mean_hat, scale=sd_hat)/(0.5*(n**2)))
			
		sd_dgp = sd_dgp * np.sqrt(abs((inv(S) * G * inv(S))[v, v]))
	
	# bandwidth
	h = np.zeros((ng, 1))
	for j in range(ng):
		if v>0:
			opt_f = lambda a: np.power(a, 2*p+2-2*v) * (bias_dgp[j, 0] + a * bias_dgp[j, 1])**2 + sd_dgp[j,0]**2 / np.power(a, 2*v - 1)
		else:
			opt_f = lambda a: np.power(a, 2*p+2-2*v) * (bias_dgp[j, 0] + a * bias_dgp[j, 1])**2 + sd_dgp[j,0]**2 / a
		h[j] = optimize.minimize_scalar(opt_f, bounds=(0, max(data)-min(data)), method='bounded').x
		
	
	for j in range(ng):
		if np.isnan(h[j])==True:
			h[j] = np.sort(abs(data-grid[j]))[min(n, max(nLocalMin, 20+p+1))]
			
		if regularize==True:
			if nLocalMin>0:
				h[j] = max(h[j], np.sort(abs(data-grid[j]))[min(n, nLocalMin)-1])
			if nUniqueMin>0:
				h[j] = max(h[j], np.sort(abs(dataUnique-grid[j]))[min(nUnique, nUniqueMin)-1])
			
			h[j] = min(h[j], max(abs(dataUnique-grid[j])))
			
	if stdVar==True:
		h = h*scale_temp
		
	return(h)


# MSE-IROT Bandwidth function
# matches R output
#slightly off only on optimize function
def __bw_IROT(data, grid, p, v, kernel, Cweights, Pweights, massPoints, stdVar, regularize, nLocalMin, nUniqueMin):
	n = len(data)
	ng = len(grid)
	
	dataUnique = np.unique(data)
	nUnique = len(dataUnique)
	
	if stdVar==True:
		center_temp = stat.mean(data)
		scale_temp = stat.stdev(data)
		data = (data-center_temp)/scale_temp
		dataUnique = (dataUnique-center_temp)/scale_temp
		grid = (grid-center_temp)/scale_temp
		
	#estimate normal reference model
	mean_hat = sum(Cweights*Pweights*data)/sum(Cweights*Pweights)
	sd_hat = np.sqrt(sum(Cweights*Pweights*(data-mean_hat)**2)/sum(Cweights*Pweights))
	
	#normal quantities
	temp_1 = sympy.exp(-(x-mean_hat)**2/(2*sd_hat**2))/sympy.sqrt(2*math.pi*sd_hat**2)
	temp_2 = sympy.diff(temp_1, x)
	temp_3 = sympy.exp(-(x-mean_hat)**2/(2*sd_hat**2))/sympy.sqrt(2*math.pi*sd_hat**2)
	temp_4 = sympy.exp(-(x-mean_hat)**2/(2*sd_hat**2))/sympy.sqrt(2*math.pi*sd_hat**2)
	
	j = p+1
	while j>1:
		temp_3 = sympy.diff(temp_3, x)
		j = j-1
	
	j=p+2
	while j >1:
		temp_4 = sympy.diff(temp_4, x)
		j = j-1
		
	#bias estimate, no rate added, DGP constant
	bias_dgp = np.zeros((ng, 2))
	for j in range(ng):
		bias_dgp[j, 0] = temp_3.evalf(subs={x:grid[j]})/math.factorial(p+1) * math.factorial(v)
		bias_dgp[j, 1] = temp_4.evalf(subs={x:grid[j]})/math.factorial(p+2) * math.factorial(v) + bias_dgp[j,0] * temp_2.evalf(subs={x:grid[j]})/temp_1.evalf(subs={x:grid[j]})
	
	#bias estimate, no rate added, kernel constant
	S = __Sgenerate(p=p, low=-1, up=1, kernel=kernel)
	C1 = __Cgenerate(k=p+1, p=p, low=-1, up=1, kernel=kernel)
	C2 = __Cgenerate(k=p+2, p=p, low=-1, up=1, kernel=kernel)
	G = __Ggenerate(p=p, low=-1, up=1, kernel=kernel)
	S2 = __Tgenerate(p=p, low=-1, up=1, kernel=kernel)
	bias_dgp[:,0] = bias_dgp[:,0] * (np.matmul(inv(S), C1)[v,:])
	bias_dgp[:,1] = bias_dgp[:,1] * (np.matmul(inv(S), C2)[v,:])
	
	#variance estimate, sample size added
	sd_dgp = np.zeros((ng, 1))
	if v>0:
		for j in range(ng):
			temp_1eval = np.array(temp_1.evalf(subs={x:grid[j]}), dtype=np.float64)
			sd_dgp[j, 0] = math.factorial(v) * np.sqrt(temp_1eval/n)
			
		sd_dgp = sd_dgp * np.sqrt(abs((np.matmul(inv(S), np.matmul(G, inv(S))))[v, v]))
		
	else:
		for j in range(ng):
			#this comes from a higher-order variance expansion. See Lemma 4 in the Appendix of Cattaneo, Jansson and Ma (2019a)
			sd_dgp[j, 0] = np.sqrt(norm.cdf(grid[j], loc=mean_hat, scale=sd_hat) * (1-norm.cdf(grid[j], loc=mean_hat, scale=sd_hat))/norm.pdf(grid[j], loc=mean_hat, scale=sd_hat)/(0.5*(n**2)))
			
		sd_dgp = sd_dgp * np.sqrt(abs((np.matmul(inv(S), np.matmul(S2, inv(S))))[v, v]))
		
	# bandwidth
	if v>0:
		opt_f = lambda a: np.power(a, 2*p+2-2*v) * sum((bias_dgp[:, 0] + a * bias_dgp[:, 1])**2) + sum(sd_dgp[:,0]**2) / np.power(a, 2*v - 1)
	else:
		opt_f = lambda a: np.power(a, 2*p+2-2*v) * sum((bias_dgp[:, 0] + a * bias_dgp[:, 1])**2) + sum(sd_dgp[:,0]**2) / a
			
	h = optimize.minimize(opt_f, x0=0.5).x
	
	if np.isnan(h)==True:
		h = np.sort(abs(data-grid[j]))[min(n, max(nLocalMin, 20+p+1))]
			
	if regularize==True:
		for j in range(ng):
			if nLocalMin>0:
				h = max(h, np.sort(abs(data-grid[j]))[min(n, nLocalMin)-1])
			if nUniqueMin>0:
				h = max(h, np.sort(abs(dataUnique-grid[j]))[min(nUnique, nUniqueMin)-1])
			
	h = min(h, max(abs(max(dataUnique)-min(grid)), abs(min(dataUnique)-max(grid))))
			
	if stdVar==True:
		h = h*scale_temp
		
	return(h)
	
	
# MSE-optimal bandwidth

def __bw_MSE(data, grid, p, v, kernel, Cweights, Pweights, massPoints, stdVar, regularize, nLocalMin, nUniqueMin):
	ii = np.argsort(data)
	data = data[ii]
	Cweights = Cweights[ii]
	Pweights = Pweights[ii]
	n = len(data)
	ng = len(grid)
	
	dataUnique = __lpdensityUnique(data)
	freqUnique = dataUnique["freq"]
	indexUnique = dataUnique["index"]
	dataUnique = dataUnique["unique"]
	nUnique = len(dataUnique)
	
	if stdVar==True:
		center_temp = stat.mean(data)
		scale_temp = stat.stdev(data)
		data = (data-center_temp)/scale_temp
		dataUnique = (dataUnique-center_temp)/scale_temp
		grid = (grid-center_temp)/scale_temp
	
	# whether considering mass points when constructing the empirical distribution function
	if massPoints == True:
		Fn = np.repeat((np.cumsum(Cweights * Pweights) / sum(Cweights * Pweights))[indexUnique], repeats = freqUnique)
	else:
		Fn = np.cumsum(Cweights * Pweights) / sum(Cweights * Pweights)
	
	weights_normal = Cweights * Pweights / sum(Cweights * Pweights) * n
	Cweights = Cweights / sum(Cweights) * n
	Pweights = Pweights / sum(Pweights) * n
	
	weights_normalUnique = np.cumsum(weights_normal)[indexUnique]
	if nUnique > 1:
		weights_normalUnique = weights_normalUnique - np.insert(weights_normalUnique[0:(nUnique-1)],0,0)
	
	CweightsUnique = np.cumsum(Cweights)[indexUnique]
	if nUnique > 1:
		CweightsUnique = CweightsUnique - np.insert(CweightsUnique[0:(nUnique-1)],0,0)
	
	PweightsUnique = np.cumsum(Pweights)[indexUnique]
	if nUnique > 1:
		PweightsUnique = PweightsUnique - np.insert(PweightsUnique[0:(nUnique-1)],0,0)
		
	# Preliminary Bandwidth estimate
	# bandwidth for preasymptotic matrics
	h1 = __bw_IROT(data=data, grid=grid, p=2, v=1, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=True, stdVar=True, regularize=True, nLocalMin=20+2+1,  nUniqueMin=20+2+1)
	#bandwidth for F_p+1
	hp1 = __bw_IROT(data=data, grid=grid, p=p+2, v=p+1, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=True, stdVar=True, regularize=True, nLocalMin=20+p+2+1,  nUniqueMin=20+p+2+1)
	
	#bandwidth fro F_p+2
	hp2 = __bw_IROT(data=data, grid=grid, p=p+3, v=p+2, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=True, stdVar=True, regularize=True, nLocalMin=20+p+3+1,  nUniqueMin=20+p+3+1)
	
	#density estimates with normalization constants
	dgp_hat = np.zeros((ng, 2))
	const_hat = np.zeros((ng, 3))
	h = np.repeat(np.nan, ng)
	
	for j in range(ng):
		#estimate F_p+2
		index_temp = abs(data-grid[j]) <= hp2
		Xh_temp = pd.DataFrame(data[index_temp]-grid[j])/hp2
		Xh_p_temp = Xh_temp[np.repeat(Xh_temp.columns, p+3+1)]
		Xh_p_temp = Xh_p_temp.apply(lambda x: np.power(x, range(p+3+1)), axis=1)
		
		
		#kernel function calculations
		Kh_temp = pd.DataFrame(__kernel_fn(Xh_temp, hp2, kernel))
		Kh_temp = Kh_temp.mul(Pweights[index_temp], axis=0)
		Y_temp = pd.DataFrame(Fn[index_temp])

		#try solving matrix inverse
		try:
			temp = (np.matmul(np.matmul(inv(np.matmul(Xh_p_temp.transpose(), Xh_p_temp.mul(Kh_temp, axis=0))), Xh_p_temp.transpose()), Y_temp.mul(Kh_temp, axis=0))).loc[p+2, 0]/np.power(hp2, p+2)
		except np.linalg.LinAlgError:
				continue
		dgp_hat[j,1] = temp
		
		

		#estimate F_p+1
		index_temp = abs(data-grid[j]) <= hp1
		Xh_temp = pd.DataFrame(data[index_temp]-grid[j])/hp1
		Xh_p_temp = Xh_temp[np.repeat(Xh_temp.columns, p+2+1)]
		Xh_p_temp = Xh_p_temp.apply(lambda x: np.power(x, range(p+2+1)), axis=1)
		
		#kernel function calculations
		Kh_temp = pd.DataFrame(__kernel_fn(Xh_temp, hp1, kernel))
#		Kh_temp = Kh_temp.mul(Pweights[index_temp], axis=0)
		Y_temp = pd.DataFrame(Fn[index_temp])
		
		#try solving matrix inverse
		try:
			temp = (np.matmul(np.matmul(inv(np.matmul(Xh_p_temp.transpose(), Xh_p_temp.mul(Kh_temp, axis=0))), Xh_p_temp.transpose()), Y_temp.mul(Kh_temp, axis=0))).loc[p+1, 0] / (hp1**(p+1))
		except np.linalg.LinAlgError:
				continue
		
		dgp_hat[j,0] = temp
		
		#estimating matrices
		index_temp = abs(data-grid[j]) <= h1
		Xh_temp = pd.DataFrame(data[index_temp]-grid[j])/h1
		Kh_temp = pd.DataFrame(__kernel_fn(Xh_temp, h1, kernel))
		
		#Cp matrix
		if p>0:
			C_p_hat = pd.DataFrame(Xh_temp[np.repeat(Xh_temp.columns, p+1)].apply(lambda x: np.power(x, range(p+1, 2*p+1+1)), axis=1).mul(Kh_temp.mul(Pweights[index_temp], axis=0), axis=0).sum(axis=0)/n)
		else:
			C_p_hat = pd.DataFrame(Xh_temp.apply(lambda x: np.power(x, 1), axis=1).mul(Kh_temp.mul(Pweights[index_temp], axis=0), axis=0).sum(axis=0)/n)
		
		#Cp+1 matrix
		if p>0:
			C_p1_hat = pd.DataFrame(Xh_temp[np.repeat(Xh_temp.columns, p+1)].apply(lambda x: np.power(x, range(p+2, 2*p+2+1)), axis=1).mul(Kh_temp.mul(Pweights[index_temp], axis=0), axis=0).sum(axis=0)/n)
		else:
			C_p1_hat = pd.DataFrame(Xh_temp.apply(lambda x: np.power(x, 2), axis=1).mul(Kh_temp.mul(Pweights[index_temp], axis=0), axis=0).sum(axis=0)/n)

		#S matrix
		Xh_p_temp = pd.DataFrame(Xh_temp[np.repeat(Xh_temp.columns, p+1)].apply(lambda x: np.power(x, range(p+1)), axis=1))
		
		
		S_hat = np.matmul(Xh_p_temp.transpose(),Xh_p_temp.mul(Kh_temp.mul(Pweights[index_temp], axis=0), axis=0))/n
		
		try:
			S_hat_inv = inv(S_hat)
		except np.linalg.LinAlgError:
			continue
			
		#G matrix
		if v==0:
			G_hat = np.matmul(Xh_p_temp, Xh_p_temp.mul(Kh_temp.mul(Pweights[index_temp], axis=0)**2)/n)
		else:
			if massPoints==True:
				Y_temp = pd.DataFrame(Fn[indexUnique])
				Xh_temp = pd.DataFrame(dataUnique-grid[j])/h1
				Xh_p_temp = Xh_temp[np.repeat(Xh_temp.columns, p+1)]
				Xh_p_temp = Xh_p_temp.apply(lambda x: np.power(x, range(p+1)), axis=1)
				if kernel == "triangular":
					Kh_temp = (1-abs(Xh_temp))/h1
					Kh_temp = Kh_temp.mul(index_temp[indexUnique], axis=0)
				elif kernel == "uniform":
					Kh_temp = index_temp[indexUnique] * 0.5/h1
				else:
					Kh_temp = 0.75*(1-np.power(Xh_temp,2))/h1
					Kh_temp = Kh_temp.mul(index_temp[indexUnique], axis=0)
				
				Xh_p_Kh_temp = Xh_p_temp.mul(Kh_temp, axis=0)
				Xh_p_Kh_Pweights_temp = Xh_p_Kh_temp.mul(PweightsUnique, axis=0)
				
				F_Xh_p_Kh_temp = np.matmul(Y_temp.transpose(), Xh_p_Kh_Pweights_temp)/n

				G = np.zeros((n,  len(Xh_p_Kh_Pweights_temp.columns)))
				for jj in range(len(Xh_p_Kh_Pweights_temp.columns)):
					G[:,jj] = (np.repeat((np.cumsum(Xh_p_Kh_Pweights_temp.iloc[::-1, jj])/n)[::-1], freqUnique) - F_Xh_p_Kh_temp.loc[0, jj]) * weights_normal
				
				G = pd.DataFrame(G)
				G_hat = np.matmul(G.transpose(), G)/n
				
			else:
				Y_temp = pd.DataFrame(Fn)
				Xh_temp = pd.DataFrame(data-grid[j])/h1
				Xh_p_temp = Xh_temp[np.repeat(Xh_temp.columns, p+1)]
				Xh_p_temp = Xh_p_temp.apply(lambda x: np.power(x, range(p+1)), axis=1)
				Kh_temp = __kernel_fn(Xh_temp, h1, kernel)
				
				Xh_p_Kh_temp = Xh_p_temp.mul(Kh_temp, axis=0)
				Xh_p_Kh_Pweights_temp = Xh_p_Kh_temp.mul(Pweights, axis=0)
				
				F_Xh_p_Kh_temp = np.matmul(Y_temp.transpose(), Xh_p_Kh_Pweights_temp)/n
				
				
				G = np.zeros((n,  len(Xh_p_Kh_Pweights_temp.columns)))
				for jj in range(len(Xh_p_Kh_Pweights_temp.columns)):
					G[:,jj] = (np.repeat(np.cumsum(Xh_p_Kh_Pweights_temp.iloc[:, jj]/n), freqUnique) - F_Xh_p_Kh_temp.loc[0, jj]) * weights_normal
				
				G = pd.DataFrame(G)
				G_hat = (np.matmul(G.transpose(), G))/n
				
		#Constants
		const_hat[j, 0] = math.factorial(v) * (np.matmul(S_hat_inv, C_p_hat)).loc[v, 0]
		const_hat[j, 1] = math.factorial(v) * (np.matmul(S_hat_inv, C_p1_hat)).loc[v, 0]
		
		if v>0:
			const_hat[j, 2] = math.factorial(v) * np.sqrt(abs(np.matmul(S_hat_inv, np.matmul(G_hat, S_hat_inv)).loc[v,v]) / (n*h1))
		else:
			temp_ii = min(max(mean(data <= grid[j]), 1/n), 1 - 1/n )
			const_hat[j, 2] = math.factorial(v) * np.sqrt(abs(np.matmul(S_hat_inv, np.matmul(G_hat, S_hat_inv)).loc[v,v] / (0.5*n**2)*h1 * temp_ii*(1-temp_ii)))

		#optimal bandwidth
		if v>0:
			opt_f = lambda a: np.power(a, 2*p+2-2*v) * (dgp_hat[j, 0]*const_hat[j,0] + a * dgp_hat[j, 1]*const_hat[j,1])**2 + const_hat[j,2]**2 / np.power(a, 2*v - 1)
		else:
			opt_f = lambda a: np.power(a, 2*p+2-2*v) * (dgp_hat[j, 0]*const_hat[j,0] + a * dgp_hat[j, 1]*const_hat[j,1])**2 + const_hat[j,2]**2 / a
		
		h[j] = optimize.minimize_scalar(opt_f, bounds=(0, max(data)-min(data)), method='bounded').x
	

	for j in range(ng):
		if np.isnan(h[j])==True:
			h[j] = np.sort(abs(data-grid[j]))[min(n, max(nLocalMin, 20+p+1))]
			
		if regularize==True:
			if nLocalMin>0:
				h[j] = max(h[j], np.sort(abs(data-grid[j]))[min(n-1, nLocalMin-1)])
			if nUniqueMin>0:
				h[j] = max(h[j], sorted(abs(dataUnique-grid[j]))[min(nUnique-1, nUniqueMin-1)])
			
			h[j] = min(h[j], max(abs(dataUnique-grid[j])))
			
	if stdVar==True:
		h = h*scale_temp
		
	return(h)

#IMSE optimal bandwidth

def __bw_IMSE(data, grid, p, v, kernel, Cweights, Pweights, massPoints, stdVar, regularize, nLocalMin, nUniqueMin):
	ii = np.argsort(data)
	data = data[ii]
#	Cweights = Cweights.reset_index(drop=True)
	Cweights = Cweights[ii]
	Pweights = Pweights[ii]
	n = len(data)
	ng = len(grid)
	
	dataUnique = __lpdensityUnique(data)
	freqUnique = dataUnique["freq"]
	indexUnique = dataUnique["index"]
	dataUnique = dataUnique["unique"]
	nUnique = len(dataUnique)
	
	if stdVar==True:
		center_temp = stat.mean(data)
		scale_temp = stat.stdev(data)
		data = (data-center_temp)/scale_temp
		dataUnique = (dataUnique-center_temp)/scale_temp
		grid = (grid-center_temp)/scale_temp
	
	# whether considering mass points when constructing the empirical distribution function
	if massPoints == True:
		Fn = np.repeat((np.cumsum(Cweights * Pweights) / sum(Cweights * Pweights))[indexUnique], repeats = freqUnique)
	else:
		Fn = np.cumsum(Cweights * Pweights) / sum(Cweights * Pweights)
		
	weights_normal = Cweights * Pweights / sum(Cweights * Pweights) * n
	Cweights = Cweights / sum(Cweights) * n
	Pweights = Pweights / sum(Pweights) * n
	
	weights_normalUnique = np.cumsum(weights_normal)[indexUnique]
	if nUnique > 1:
		weights_normalUnique = weights_normalUnique - np.insert(weights_normalUnique[:-1],0,0)
	
	CweightsUnique = np.cumsum(Cweights)[indexUnique]
	if nUnique > 1:
		CweightsUnique = CweightsUnique - np.insert(CweightsUnique[:-1],0,0)
	
	PweightsUnique = np.cumsum(Pweights)[indexUnique]
	if nUnique > 1:
		PweightsUnique = PweightsUnique - np.insert(PweightsUnique[:-1],0,0)
		
	# Preliminary Bandwidth estimate
	# bandwidth for preasymptotic matrics
	h1 = __bw_IROT(data=data, grid=grid, p=2, v=1, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=True, stdVar=True, regularize=True, nLocalMin=20+2+1,  nUniqueMin=20+2+1)
	
	#bandwidth for F_p+1
	hp1 = __bw_IROT(data=data, grid=grid, p=p+2, v=p+1, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=True, stdVar=True, regularize=True, nLocalMin=20+p+2+1,  nUniqueMin=20+p+2+1)

	#bandwidth fro F_p+2
	hp2 = __bw_IROT(data=data, grid=grid, p=p+3, v=p+2, kernel=kernel, Cweights=Cweights, Pweights=Pweights, massPoints=True, stdVar=True, regularize=True, nLocalMin=20+p+3+1,  nUniqueMin=20+p+3+1)

	#density estimates with normalization constants
	dgp_hat = np.zeros((ng, 2))
	const_hat = np.zeros((ng, 3))
	
	for j in range(ng):
		#estimate F_p+2
		index_temp = abs(data-grid[j]) <= hp2
		Xh_temp = pd.DataFrame(data[index_temp]-grid[j])/hp2
		Xh_p_temp = Xh_temp[np.repeat(Xh_temp.columns, p+3+1)]
		Xh_p_temp = Xh_p_temp.apply(lambda x: np.power(x, range(p+3+1)), axis=1)
		
		
		#kernel function calculations
		Kh_temp = pd.DataFrame(__kernel_fn(Xh_temp, hp2, kernel))
		Kh_temp = Kh_temp.mul(Pweights[index_temp], axis=0)
		Y_temp = pd.DataFrame(Fn[index_temp])

		#try solving matrix inverse
		try:
			temp = (np.matmul(np.matmul(inv(np.matmul(Xh_p_temp.transpose(), Xh_p_temp.mul(Kh_temp, axis=0))), Xh_p_temp.transpose()), Y_temp.mul(Kh_temp, axis=0))).loc[p+2, 0]/np.power(hp2, p+2)
		except np.linalg.LinAlgError:
				continue
		dgp_hat[j,1] = temp
		
		

		#estimate F_p+1
		index_temp = abs(data-grid[j]) <= hp1
		Xh_temp = pd.DataFrame(data[index_temp]-grid[j])/hp1
		Xh_p_temp = Xh_temp[np.repeat(Xh_temp.columns, p+2+1)]
		Xh_p_temp = Xh_p_temp.apply(lambda x: np.power(x, range(p+2+1)), axis=1)
		
		#kernel function calculations
		Kh_temp = pd.DataFrame(__kernel_fn(Xh_temp, hp1, kernel))
#		Kh_temp = Kh_temp.mul(Pweights[index_temp], axis=0)
		Y_temp = pd.DataFrame(Fn[index_temp])
		
		#try solving matrix inverse
		try:
			temp = (np.matmul(np.matmul(inv(np.matmul(Xh_p_temp.transpose(), Xh_p_temp.mul(Kh_temp, axis=0))), Xh_p_temp.transpose()), Y_temp.mul(Kh_temp, axis=0))).loc[p+1, 0] / (hp1**(p+1))
		except np.linalg.LinAlgError:
				continue
		
		dgp_hat[j,0] = temp
		
		#estimating matrices
		index_temp = abs(data-grid[j]) <= h1
		Xh_temp = pd.DataFrame(data[index_temp]-grid[j])/h1
		Kh_temp = pd.DataFrame(__kernel_fn(Xh_temp, h1, kernel))
		
		#Cp matrix
		if p>0:
			C_p_hat = pd.DataFrame(Xh_temp[np.repeat(Xh_temp.columns, p+1)].apply(lambda x: np.power(x, range(p+1, 2*p+1+1)), axis=1).mul(Kh_temp.mul(Pweights[index_temp], axis=0), axis=0).sum(axis=0)/n)
		else:
			C_p_hat = pd.DataFrame(Xh_temp.apply(lambda x: np.power(x, 1), axis=1).mul(Kh_temp.mul(Pweights[index_temp], axis=0), axis=0).sum(axis=0)/n)
		
		#Cp+1 matrix
		if p>0:
			C_p1_hat = pd.DataFrame(Xh_temp[np.repeat(Xh_temp.columns, p+1)].apply(lambda x: np.power(x, range(p+2, 2*p+2+1)), axis=1).mul(Kh_temp.mul(Pweights[index_temp], axis=0), axis=0).sum(axis=0)/n)
		else:
			C_p1_hat = pd.DataFrame(Xh_temp.apply(lambda x: np.power(x, 2), axis=1).mul(Kh_temp.mul(Pweights[index_temp], axis=0), axis=0).sum(axis=0)/n)

		#S matrix
		Xh_p_temp = pd.DataFrame(Xh_temp[np.repeat(Xh_temp.columns, p+1)].apply(lambda x: np.power(x, range(p+1)), axis=1))
		
		
		S_hat = np.matmul(Xh_p_temp.transpose(),Xh_p_temp.mul(Kh_temp.mul(Pweights[index_temp], axis=0), axis=0))/n
		
		try:
			S_hat_inv = inv(S_hat)
		except np.linalg.LinAlgError:
			continue
			
		#G matrix
		if v==0:
			G_hat = np.matmul(Xh_p_temp, Xh_p_temp.mul(Kh_temp.mul(Pweights[index_temp], axis=0)**2)/n)
		else:
			if massPoints==True:
				Y_temp = pd.DataFrame(Fn[indexUnique])
				Xh_temp = pd.DataFrame(dataUnique-grid[j])/h1
				Xh_p_temp = Xh_temp[np.repeat(Xh_temp.columns, p+1)]
				Xh_p_temp = Xh_p_temp.apply(lambda x: np.power(x, range(p+1)), axis=1)

				if kernel == "triangular":
					Kh_temp = (1-abs(Xh_temp))/h1
					Kh_temp = Kh_temp.mul(index_temp[indexUnique], axis=0)
				elif kernel == "uniform":
					Kh_temp = index_temp[indexUnique] * 0.5/h1
				else:
					Kh_temp = 0.75*(1-np.power(Xh_temp,2))/h1
					Kh_temp = Kh_temp.mul(index_temp[indexUnique], axis=0)
				
				Xh_p_Kh_temp = Xh_p_temp.mul(Kh_temp, axis=0)
				Xh_p_Kh_Pweights_temp = Xh_p_Kh_temp.mul(PweightsUnique, axis=0)
				
				F_Xh_p_Kh_temp = np.matmul(Y_temp.transpose(), Xh_p_Kh_Pweights_temp)/n

				G = np.zeros((n,  len(Xh_p_Kh_Pweights_temp.columns)))
				for jj in range(len(Xh_p_Kh_Pweights_temp.columns)):
					G[:,jj] = (np.repeat((np.cumsum(Xh_p_Kh_Pweights_temp.iloc[::-1, jj])/n)[::-1], freqUnique) - F_Xh_p_Kh_temp.loc[0, jj]) * weights_normal
				
				G = pd.DataFrame(G)
				G_hat = np.matmul(G.transpose(), G)/n
				
			else:
				Y_temp = pd.DataFrame(Fn)
				Xh_temp = pd.DataFrame(data-grid[j])/h1
				Xh_p_temp = Xh_temp[np.repeat(Xh_temp.columns, p+1)]
				Xh_p_temp = Xh_p_temp.apply(lambda x: np.power(x, range(p+1)), axis=1)
				Kh_temp = __kernel_fn(Xh_temp, h1, kernel)
				
				Xh_p_Kh_temp = Xh_p_temp.mul(Kh_temp, axis=0)
				Xh_p_Kh_Pweights_temp = Xh_p_Kh_temp.mul(Pweights, axis=0)
				
				F_Xh_p_Kh_temp = np.matmul(Y_temp.transpose(), Xh_p_Kh_Pweights_temp)/n
				
				
				G = np.zeros((n,  len(Xh_p_Kh_Pweights_temp.columns)))
				for jj in range(len(Xh_p_Kh_Pweights_temp.columns)):
					G[:,jj] = (np.cumsum(Xh_p_Kh_Pweights_temp.iloc[:, jj]/n) - F_Xh_p_Kh_temp.loc[0, jj]) * weights_normal
				
				G = pd.DataFrame(G)
				G_hat = (np.matmul(G.transpose(), G))/n
				
		#Constants
		const_hat[j, 0] = math.factorial(v) * (np.matmul(S_hat_inv, C_p_hat)).loc[v, 0]
		const_hat[j, 1] = math.factorial(v) * (np.matmul(S_hat_inv, C_p1_hat)).loc[v, 0]
		
		if v>0:
			const_hat[j, 2] = math.factorial(v) * np.sqrt(abs(np.matmul(S_hat_inv, np.matmul(G_hat, S_hat_inv)).loc[v,v]) / (n*h1))
		else:
			temp_ii = min(max(mean(data <= grid[j]), 1/n), 1 - 1/n )
			const_hat[j, 2] = math.factorial(v) * np.sqrt(abs(np.matmul(S_hat_inv, np.matmul(G_hat, S_hat_inv)).loc[v,v] / (0.5*n**2)*h1 * temp_ii*(1-temp_ii)))
	

	#optimal bandwidth
	if v>0:
		opt_f = lambda a: np.power(a, 2*p+2-2*v) * sum((dgp_hat[:, 0]* const_hat[:, 0] + a * dgp_hat[:, 1]* const_hat[:, 1])**2) + sum(const_hat[:,2]**2) / np.power(a, 2*v - 1)
	else:
		opt_f = lambda a: np.power(a, 2*p+2-2*v) * sum((dgp_hat[:, 0]* const_hat[:, 0] + a * dgp_hat[:, 1]* const_hat[:, 1])**2) + sum(const_hat[:,2]**2) / a
			
	h = optimize.minimize_scalar(opt_f, bounds=(0, max(data)-min(data)), method='bounded').x
	
	
	if np.isnan(h)==True:
		h = max(h, np.sort(abs(data-grid[j]))[min(n, max(nLocalMin, 20+p+1))])
	
	if regularize==True:
		for j in range(ng):
			if nLocalMin>0:
				h = max(h, np.sort(abs(data-grid[j]))[min(n-1, nLocalMin-1)])
			if nUniqueMin>0:
				h = max(h, sorted(abs(dataUnique-grid[j]))[min(nUnique-1, nUniqueMin-1)])

	h = min(h, max(abs(max(dataUnique)-min(grid)), abs(min(dataUnique)-max(grid))))
			
	if stdVar==True:
		h = h*scale_temp
		
	return(h)
